require 'rails_helper'
require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')

RSpec.describe Hermes::HousingApplication, type: :model do
  describe 'edit with change status' do
    let!(:entity) {create(:entity, siret: '00000000000001')}
    let!(:applicant) {create(:applicant, main_employer_siret: entity.siret)}
    let!(:applicant2) {create(:applicant, main_employer_siret: entity.siret)}
    let!(:applicant3) {create(:applicant, main_employer_siret: entity.siret)}

    let!(:housing_offer1) do
      create(
        :housing_offer,
        typology: 'T2',
        external_ref: 'blabli'
      )
    end
    
    let!(:housing_request1) do
      create(
        :complete_request,
        is_active: true,
        main_applicant: applicant
      )
    end
    
    let!(:housing_request2) do
      create(
        :complete_request,
        is_active: true,
        main_applicant: applicant2
      )
    end

    let!(:housing_request3) do
      create(
        :complete_request,
        is_active: true,
        main_applicant: applicant3
      )
    end
    
    let!(:housing_application1) do
      create(
        :housing_application,
        housing_offer_id: housing_offer1.id,
        housing_request_id: housing_request1.id,
        status: 'TRANSMIS BAILLEUR'
      )
    end
    let!(:housing_application2) do
      create(
        :housing_application,
        housing_offer_id: housing_offer1.id,
        housing_request_id: housing_request2.id,
        status: "A L'ETUDE ALS"
      )
    end
    
    let!(:housing_application3) do
      create(
        :housing_application,
        housing_offer_id: housing_offer1.id,
        housing_request_id: housing_request3.id,
        status: 'TRANSMIS BAILLEUR'
      )
    end
    
    it 'triggers sopra flux' do
      # allow(housing_application1).to receive(:update_status).and_call_original
      allow(Hermes::BoClient::UpdateStatusTask).to receive_message_chain(:new, :execute)
      housing_application1.update(status: 'RETENUE EN CAL')
      
      # expect(housing_application1).to have_received(:update_status).once
      # expect(Hermes::BoClient::UpdateStatusTask.new).to have_received(:execute).once
    end
    it 'triggers sopra flux' do
      # allow(housing_application1).to receive(:update_status).and_call_original
      allow(Hermes::BoClient::UpdateStatusTask).to receive_message_chain(:new, :execute)
      housing_application1.update!(status: 'LOGE')
  
      expect(housing_application1.housing_request.reload.request_status).to eq('SATISFAITE')
      expect(housing_application1.housing_request.reload.entity_status).to eq('SALARIE LOGE')
    
      expect(housing_application1.housing_offer.reload.offer_status).to eq('LOGEMENT PLACE')
      # expect(Hermes::BoClient::UpdateStatusTask.new).to have_received(:execute).once
    end


    it 'increment is visible in housing_request_serializer' do
      json = Hermes::HousingRequestSerializer.new(
        housing_request1.reload,
        # scope: scope,
        root: false
      ).to_json
      expect(JSON.parse(json)['applications_nb']).to eq(1)
    end
  end
  
  describe 'Statues' do
    let(:applicant) {create(:applicant)}
    let(:housing_offer1) {create(:housing_offer, offer_status: 'LOGEMENT PUBLIE', typology: 'T1')}
    let(:housing_request) {create(:housing_request, is_active: true, main_applicant: applicant)}
    let(:housing_application1) do
      create(
        :housing_application,
        housing_offer_id: housing_offer1.id,
        housing_request_id: housing_request.id,
        status: 'TRANSMIS BAILLEUR'
      )
    end
    
    it 'trigger acceptation' do
      housing_application1.update!(status: 'RETENUE EN CAL', visit_status: 'VISITE EFFECTUEE')
      applicant.reload
      expect(applicant.completed_requests_nb).to eq(1)
    end
  end
  describe 'Cancelation' do
    let!(:canceler_applicant) {create(:applicant)}
    let!(:housing_offer1) {create(:housing_offer, offer_status: 'LOGEMENT PUBLIE', typology: 'T1')}
    let!(:housing_offer2) {create(:housing_offer, offer_status: 'LOGEMENT PUBLIE', typology: 'T1')}
    let!(:housing_request) {create(:housing_request, is_active: true, main_applicant: canceler_applicant)}
    let(:housing_application1) do
      create(
        :housing_application,
        housing_offer_id: housing_offer1.id,
        housing_request_id: housing_request.id,
        status: 'TRANSMIS BAILLEUR',
        rank: 2
      )
    end
    let(:housing_application2) do
      create(
        :housing_application,
        housing_offer_id: housing_offer2.id,
        housing_request_id: housing_request.id,
        status: 'TRANSMIS BAILLEUR'
      )
    end
    
    before do
      [housing_application1, housing_application2].each(&:reload)
    end
    
    it 'trigger cancelation' do
      expect(housing_request.reload.active_application_id).to eq(housing_application2.id.to_s)
      expect(housing_request.reload.active_application_housing_offer_ids.map { |a| a[:housing_offer_id]}).to match_array([housing_offer1.id.to_s, housing_offer2.id.to_s])

      housing_application1.update(visit_status: 'DESISTEMENT APRES VISITE')
      housing_application2.update(status: 'DESISTEMENT')
      
      [housing_application1, housing_application2].each do |app|
        expect(app.reload.is_active).to be(false)
      end
      
      expect(housing_request.reload.applications_nb).to eq(2)
      expect(housing_request.reload.active_application_id).to eq(nil)
      expect(housing_request.reload.active_application_housing_offer_ids).to be_empty
    end
  end
  
  describe 'Hooks' do
    let!(:lessor_user) {create(:user, first_name: 'Jaques', last_name: 'Dupont', email: 'jaques.dupont@gmail.com', roles: ['bailleur'])}
    let!(:lessor) {create(:lessor, user_id: lessor_user.id.to_s)}
    let!(:housing_offer) {create(:housing_offer, lessor: lessor)}
    let!(:applicant) {create(:applicant)}
    let!(:applicant2) {create(:applicant)}
    let!(:housing_request) {create(:housing_request, main_applicant_id: applicant.id, is_active: true, coapplicants: [{id: applicant2.id, rel_type: 'C'}], als_id: '00000001')}
    let!(:housing_application) {create(:housing_application, housing_request: housing_request, housing_offer: housing_offer)}
    
    before do
      applicant.reload
      applicant2.reload
    end
    
    it 'setup lessor as readers' do
      expect(housing_application.readers).to include(lessor_user.id.to_s)
      expect(housing_request.readers).to include(lessor_user.id.to_s)
      expect(applicant.readers).to include(lessor_user.id.to_s)
      expect(applicant2.readers).to include(lessor_user.id.to_s)
    end
    
    it 'remove lessor as readers' do
      housing_application.update!(status: 'ANNULEE BAILLEUR')
      applicant.reload
      applicant2.reload
      expect(housing_application.reload.readers).not_to include(lessor_user.id.to_s)
      expect(housing_request.reload.readers).not_to include(lessor_user.id.to_s)
      expect(applicant.readers).not_to include(lessor_user.id.to_s)
      expect(applicant2.readers).not_to include(lessor_user.id.to_s)
    end

    it 'should use als_id from request if not exist on application' do
      expect(housing_application.als_id).to eq(housing_request.als_id)
    end
  end
  
  describe "add readers and writers from offer" do
    let(:user) {create :user}
    let(:applicant) {create :applicant, user: user}
    let(:request) {create :housing_request, main_applicant: applicant}
    let(:reader_offer1) {create :user}
    let(:writer_offer1) {create :user}
    let(:reader_offer2) {create :user}
    
    let(:offer1) {create :housing_offer, readers: [reader_offer1.id.to_s], writers: [writer_offer1.id.to_s]}
    let(:offer2) {create :housing_offer, readers: [reader_offer2.id.to_s]}
    let(:application) {create :housing_application, housing_request: request, housing_offer: offer1}
    
    it 'report readers and writers after creation' do
      expect(application.readable_by?(reader_offer1)).to eq(true)
      expect(application.readable_by?(reader_offer2)).to eq(false)
      expect(application.writable_by?(writer_offer1)).to eq(true)
    end
    
    it 'report readers and writers after udapted offer_id' do
      application.housing_offer = offer2
      application.save
      expect(application.readable_by?(reader_offer1)).to eq(false)
      expect(application.readable_by?(reader_offer2)).to eq(true)
      expect(application.writable_by?(writer_offer1)).to eq(false)
    end
    
    it 'report readers and writers after offer readers change' do
      offer1.readers = [writer_offer1.id.to_s]
      offer1.writers = []
      offer1.save
      expect(application.readable_by?(reader_offer1)).to eq(false)
      expect(application.readable_by?(reader_offer2)).to eq(false)
      expect(application.readable_by?(writer_offer1)).to eq(true)
      expect(application.writable_by?(writer_offer1)).to eq(false)
    end
  
  
  end

  describe "get user_id from request" do
    let!(:user) { create(:user)}
    let!(:applicant) { create(:applicant)}
    let!(:o) { create(:housing_offer) }
    
    let!(:hr) do
      create(
        :housing_request,
        user: user,
        is_active: true,
        main_applicant: applicant
      )
    end

    let!(:ha) { create(:housing_application, housing_request: hr, housing_offer: o)}

    it 'works' do 
      expect(ha.user_id).to eq(user.id)
      expect(ha.writable_by?(user)).to eq(true)
    end 
  end 


  describe "check_order_status" do 
    let!(:user) { create(:user)}
    let!(:applicant) { create(:applicant)}
    let!(:housing_request) { create(:housing_request, is_active: true, main_applicant: applicant) }
    let!(:offer) { create(:housing_offer) }
    let!(:initial_values) do 
      {
        status: 'RETENUE EN CAL SOUS CONDITIONS',
        rank: 1,
        last_status_update_date: Date.new(2020,1,1),
        date_event: Date.new(2020,1,2)
      }
    end 
    let!(:ha) { create(:housing_application, initial_values.merge(is_active: true, housing_request: housing_request, housing_offer: offer))} 

    it 'cancel status when not in order' do
      ha.update(
        status: 'VISITE PROPOSEE',
        date_event: Date.new(2020,1,3),
        rank: 10,
        last_status_update_date: Date.new(2020,1,4),
      )
      ha.reload
      initial_values.keys.each do |key|
        expect(ha[key]).to eq(initial_values[key])
      end 
      
    end
    
    it 'not do anything when not in mapping' do 
      new_values = {
        status: 'ANNULEE',
        date_event: Date.new(2020,1,3),
        rank: 10,
        last_status_update_date: Date.new(2020,1,4),
      }
      ha.update(new_values)
      ha.reload
      new_values.keys.each do |key|
        expect(ha[key]).to eq(new_values[key])
      end 
    end 

    it 'work properly when in order' do 
      new_values = {
        status: 'LOGE',
        date_event: Date.new(2020,1,3),
        rank: 10,
        last_status_update_date: Date.new(2020,1,4),
      }
      ha.update(new_values)
      ha.reload
      new_values.keys.each do |key|
        expect(ha[key]).to eq(new_values[key])
      end 
    end 

  end 

  describe "notify_applicant" do 
    let(:user) {create(:user)}
    let(:applicant) {create(:applicant, :with_eligible_offers, offers_count: 3, user: user)}
    let(:offer) { applicant.first_active_request.eligibles_offers.first}
    let(:ha) do
      create(
        :housing_application,
        user: user,
        housing_offer_id: offer.id,
        housing_request_id: applicant.first_active_request.id,
        status: 'TRANSMIS BAILLEUR'
      )
    end

    before do 
      create_mail_template_default()
      applicant.first_active_request.eligibles_offers.each do |o|
        o.update(
          district: 'Marseille',
        )
      end 
      offer.update(
        external_ref: 'ALS_01',
        district: 'Paris',
      )
    end
    
    it 'not send application_refusal' do 
      expect(applicant.first_active_request.reload.eligibles_offers.count).to eq(3)
      ha.update(status: 'REFUSEE EN CAL')
      last_notif_job = enqueued_jobs.select{|job| job[:job] == AsyncNotifierJob }.last
      expect(last_notif_job[:args][0]['key']).not_to eq('user.application_refusal')
    end 

    ['IV01', 'IV02', 'IV05'].each do |motive_code|
      it "don't send application_not_selected when #{motive_code}" do 
        ha.update(status: 'NON RETENUE', motive_code: motive_code)
        last_notif_job = enqueued_jobs.select{|job| job[:job] == AsyncNotifierJob }.last
        expect(last_notif_job[:args][0]['key']).not_to eq('user.application_not_selected')
      end 
    end 

    it 'send application_not_selected' do 
      expect(applicant.first_active_request.reload.eligibles_offers.count).to eq(3)
      ha.update(status: 'NON RETENUE')
      last_notif_job = enqueued_jobs.select{|job| job[:job] == AsyncNotifierJob }.last
      expect(last_notif_job[:args][0]['key']).to eq('user.application_not_selected')

      last_notif_job[:job].perform_now(*deserialize_argument(last_notif_job[:args]))
    end 

    it 'send application_offer_taken_back' do 
      expect(applicant.first_active_request.reload.eligibles_offers.count).to eq(3)

      ha.update(status: 'REPRIS BAILLEUR')
      last_notif_job = enqueued_jobs.select{|job| job[:job] == AsyncNotifierJob }.last
      expect(last_notif_job[:args][0]['key']).to eq('user.application_offer_taken_back')

      last_notif_job[:job].perform_now(*deserialize_argument(last_notif_job[:args]))
    end 
  end   

  describe 'report target_offer_reserving_business' do 
    let!(:rh) { create(:reservation_holder, name: 'ALS', siren: ['123456789'])}
    let!(:offer) { create(:housing_offer, reserving_siren: rh.siren.first)}
    let!(:hr) {create(:active_request)}
    let!(:app) {create(:housing_application, housing_request: hr, housing_offer: offer, status: 'EN ATTENTE')}

    it 'works when not active' do 
      expect(hr.reload.target_offer_reserving_businesses.first[:reserving_siren]).to eq('ALS')

      app.update(status: 'ANNULEE')
      expect(hr.reload.target_offer_reserving_businesses).to be_empty
    end 

    it 'works when LOGE' do 
      app.update(status: 'LOGE')
      expect(hr.reload.target_offer_reserving_businesses.first[:reserving_siren]).to eq('ALS')
    end 
  end 

  describe 'not send flux when not matching status' do 
    let!(:offer) { create(:housing_offer)}
    let!(:hr) {create(:active_request)}
    let!(:ha) {create(:housing_application, housing_request: hr, housing_offer: offer, status: 'EN ATTENTE')}

    it 'works' do 
      expect(Hermes::Statuses::HousingApplication.sopra_code(ha.status)).to eq(nil)
      expect(enqueued_jobs.select{|job| job[:job] == ExecuteSopraTaskJob }.length).to eq(0)
      ha.update(status: 'INSCRIPTION CAL')
      expect(Hermes::Statuses::HousingApplication.sopra_code(ha.status)).to eq('INSCAL')
      expect(enqueued_jobs.select{|job| job[:job] == ExecuteSopraTaskJob }.length).to eq(1)
    end 
  end 

  describe 'only one application active per offer' do 
    let!(:offer) { create(:housing_offer)}
    let!(:offer2) { create(:housing_offer)}
    let!(:hr) {create(:complete_request)}
    subject { create(:housing_application, housing_request: hr, housing_offer: offer, status: 'EN ATTENTE')}

    it 'works' do
      expect { subject }.not_to raise_error
    end

    it 'when one active application' do
      create(:housing_application, housing_request: hr, housing_offer: offer, status: 'EN ATTENTE', rank: 2)
      expect { subject }.to raise_error(Mongoid::Errors::Validations)
    end

    it 'when one active application' do
      app = subject 
      app.update(rank: 2)
      other_app = create(:housing_application, housing_request: hr, housing_offer: offer2, status: 'EN ATTENTE', rank: 2)
      expect { other_app.update!(housing_offer: offer) }.to raise_error(Mongoid::Errors::Validations)
    end

  end 

  describe 'can_create_application' do 
    let(:housing_request) {create(:complete_request)}
    let(:offer) {create(:housing_offer)}
    let(:offer2) {create(:housing_offer)}

    subject { create(:housing_application, housing_request: housing_request, housing_offer: offer, status: 'EN ATTENTE')}

    it 'works when no other application' do
      expect { subject }.not_to raise_error
    end 

    it 'with one active application' do
      create(:housing_application, housing_request: housing_request, housing_offer: offer2, status: 'EN ATTENTE')
      expect { subject }.to raise_error(Mongoid::Errors::Validations)
    end

    it 'with one dormant application' do
      create(:housing_application, housing_request: housing_request, housing_offer: offer2, rank: 2, status: 'EN ATTENTE')
      expect { subject }.not_to raise_error
    end 

    it 'with 2 dormant application' do
      create(:housing_application, housing_request: housing_request, housing_offer: offer, rank: 2, status: 'EN ATTENTE')
      create(:housing_application, housing_request: housing_request, housing_offer: offer2, rank: 2, status: 'EN ATTENTE')
      expect { subject }.to raise_error(Mongoid::Errors::Validations)
    end

    it 'when changing rank' do
      app = subject
      app.update!(rank: 2)
      create(:housing_application, housing_request: housing_request, housing_offer: offer2, rank: 1, status: 'EN ATTENTE')
      
      expect{ app.update!(rank: 1) }.not_to raise_error
    end 

    it 'when changing status' do
      app = subject
      app.update!(status: 'ANNULEE')
      create(:housing_application, housing_request: housing_request, housing_offer: offer2, rank: 1, status: 'EN ATTENTE')

      expect{ app.update!(status: 'EN ATTENTE') }.not_to raise_error
    end 
  end 

  describe 'update_request_dormant_application_ids' do 
    let!(:housing_request) {create(:complete_request)}
    let!(:offer) {create(:housing_offer)}
    let!(:app) {create(:housing_application, housing_request: housing_request, housing_offer: offer, status: 'EN ATTENTE')}

    it 'one active application' do
      expect(housing_request.reload.dormant_application_ids).to be_empty
      expect(housing_request.reload.active_application_id).to eq(app.id.to_s)
    end 

    it 'when canceling applicaiton' do 
      app.update!(status: 'ANNULEE')

      expect(housing_request.reload.dormant_application_ids).to be_empty
      expect(housing_request.reload.active_application_id).to be_nil
    end 

    it 'when update to rank 2' do
      app.update(rank: 2)

      expect(housing_request.reload.dormant_application_ids).to eq([app.id.to_s])
      expect(housing_request.reload.active_application_id).to be_nil
    end

    it 'when update to rank 1' do
      app.update(rank: 2)
      app.update(rank: 1)

      expect(housing_request.reload.dormant_application_ids).to be_empty
      expect(housing_request.reload.active_application_id).to eq(app.id.to_s)
    end
  end 

  describe "cancel_other_applications" do 
    let!(:housing_request) {create(:complete_request)}
    let!(:offer) {create(:housing_offer)}
    let!(:offer2) {create(:housing_offer)}
    let!(:app) {create(:housing_application, housing_request: housing_request, housing_offer: offer, status: 'EN ATTENTE', rank: 2)}
    let!(:app2) {create(:housing_application, housing_request: housing_request, housing_offer: offer2, status: 'EN ATTENTE', rank: 2)}

    it 'works' do 
      app.update(status: 'LOGE')

      expect(app2.reload.status).to eq('ANNULEE')
    end 

  end 
end
